// This code example is created for educational purpose
// by Thorsten Thormaehlen (contact: www.thormae.de).
// It is distributed without any warranty.

package de.unimarburg.shaderlightmatapp;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import android.app.Activity;
import android.content.Context;
import android.opengl.GLES20;
import android.opengl.GLSurfaceView;
import android.os.Bundle;
import android.os.SystemClock;
import android.util.Log;
import android.view.GestureDetector;
import android.view.MotionEvent;

class Renderer {

  public float t;
  public int modeVal = 1;
  public boolean loadNext = false;
  public int shaderVertId = R.raw.phongbrdfpervertex_vert;
  public int shaderFragId = R.raw.phongbrdfpervertex_frag;
  public int shaderFile = 0;
  public int vboResourceId = R.raw.knot;
  public int vboFile = 0;
  
  private String TAG = "de.unimarburg.shaderlightmatapp";
  private Context appContext;
  private enum VBOs {SceneAll, numVBOs};
  private int[] bufID = new int[VBOs.numVBOs.ordinal()];
  private int sceneVertNo = 0;
  private int progID = 0;
  private int vertID = 0;
  private int fragID = 0;
  private int vertexLoc = 0;
  private int texCoordLoc = 0;
  private int normalLoc = 0;
  private int projectionLoc = 0;
  private int modelviewLoc = 0;
  private int normalMatrixLoc = 0;
  private int modeLoc = 0;
  private float[] projection = new float[16];
  private float[] modelview = new float[16];
  
  
  private int SIZEOF_FLOAT = Float.SIZE / 8;
  
  public Renderer(Context context) {
    appContext = context;
  }
  
  public void next(GL10 unused) {
    loadNext = false;
    dispose(unused);
    reload(unused);
  }

  public void init(GL10 unused, EGLConfig config) {
    reload(unused);
  }

  public void reload(GL10 unused) {
    GLES20.glEnable(GLES20.GL_DEPTH_TEST);

    setupShaders();

    // generate a Vertex Buffer Object (VBO)
    GLES20.glGenBuffers(VBOs.numVBOs.ordinal(), bufID, 0);

    int perVertexFloats = (3 + 2 + 3);
    float data[] = loadVertexData(vboResourceId, perVertexFloats);
    sceneVertNo = data.length / perVertexFloats;

    ByteBuffer bb = ByteBuffer.allocateDirect(data.length * SIZEOF_FLOAT);
    bb.order(ByteOrder.nativeOrder());
    FloatBuffer sceneVertexFB = bb.asFloatBuffer();
    sceneVertexFB.put(data);
    sceneVertexFB.flip();

    GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, bufID[VBOs.SceneAll.ordinal()]);
    GLES20.glBufferData(GLES20.GL_ARRAY_BUFFER, sceneVertexFB.capacity()
        * SIZEOF_FLOAT, sceneVertexFB, GLES20.GL_STATIC_DRAW);
  }
  
  public void resize(GL10 unused, int width, int height) {
    GLES20.glViewport(0, 0, width, height);
    
    // this function replaces gluPerspective
    mat4Perspective(projection, 45.0f, (float)width/(float)height, 0.5f, 4.0f);
    //mat4Print(projection);
  }

  
  public void display(GL10 unused) {

    if(loadNext) next(unused);
    
    GLES20.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);

    // camera orbits in the z=1.5 plane
    // and looks at the origin
    // mat4LookAt replaces gluLookAt
    double rad = Math.PI / 180.0f * t;
    mat4LookAt(modelview,
        1.5f*(float)Math.cos(rad), 1.5f*(float)Math.sin(rad), 1.5f, // eye
        0.0f, 0.0f, 0.0f, // look at
        0.0f, 0.0f, 1.0f); // up
    
    float modelviewInv[] = new float[16];
    float normalmatrix[] = new float[16];
    mat4Invert(modelview, modelviewInv);
    mat4Transpose(modelviewInv, normalmatrix);
    
    GLES20.glUseProgram(progID); 
    
    // load the current projection and modelview matrix into the
    // corresponding UNIFORM variables of the shader
    GLES20.glUniformMatrix4fv(projectionLoc, 1, false, projection, 0);
    GLES20.glUniformMatrix4fv(modelviewLoc, 1, false, modelview, 0);
    GLES20.glUniformMatrix4fv(normalMatrixLoc, 1, false, normalmatrix, 0);
    GLES20.glUniform1i(modeLoc, modeVal);
    
    // bind scene VBO
    GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, bufID[VBOs.SceneAll.ordinal()]);

    int stride = (3+2+3)*SIZEOF_FLOAT;
    int offset = 0;

    // position
    if (vertexLoc != -1) {
      GLES20.glVertexAttribPointer(vertexLoc, 3,  GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(vertexLoc);
    }

    // texCoord
    if (texCoordLoc != -1) {
      offset = 0 + 3 * SIZEOF_FLOAT;
      GLES20.glVertexAttribPointer(texCoordLoc, 2,  GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(texCoordLoc);
    }

    // normal
    if (normalLoc != -1) {
      offset = 0 + (3 + 2) * SIZEOF_FLOAT;
      GLES20.glVertexAttribPointer(normalLoc, 3, GLES20.GL_FLOAT, false, stride,
          offset);
      GLES20.glEnableVertexAttribArray(normalLoc);
    }
    
    // render data
    GLES20.glDrawArrays(GLES20.GL_TRIANGLES, 0, sceneVertNo);

    GLES20.glFlush();
  }
  
  
  public void dispose(GL10 unused) {
    GLES20.glDeleteBuffers(VBOs.numVBOs.ordinal(), bufID, 0);
    GLES20.glDeleteProgram(progID);
    GLES20.glDeleteShader(vertID);
    GLES20.glDeleteShader(fragID);
  }
   
  public void setupShaders() {
     
    vertID = GLES20.glCreateShader(GLES20.GL_VERTEX_SHADER);
    fragID = GLES20.glCreateShader(GLES20.GL_FRAGMENT_SHADER);
     
    String vs = loadShaderSrc(shaderVertId); 
    String fs = loadShaderSrc(shaderFragId);
   
    GLES20.glShaderSource(vertID, vs);
    GLES20.glShaderSource(fragID, fs);

    // compile the shader
    GLES20.glCompileShader(vertID);
    GLES20.glCompileShader(fragID);

    // check for errors
    printShaderInfoLog(vertID);
    printShaderInfoLog(fragID);

    // create program and attach shaders
    progID = GLES20.glCreateProgram();
    GLES20.glAttachShader(progID, vertID);
    GLES20.glAttachShader(progID, fragID);

    // link the program
    GLES20.glLinkProgram(progID);
    // output error messages
    printProgramInfoLog(progID);

    // retrieve the location of the IN variables of the vertex shader
    vertexLoc = GLES20.glGetAttribLocation(progID,"inputPosition");
    texCoordLoc =  GLES20.glGetAttribLocation(progID,"inputTexCoord");
    normalLoc = GLES20.glGetAttribLocation(progID, "inputNormal");
    
    // retrieve the location of the UNIFORM variables of the vertex shader
    projectionLoc = GLES20.glGetUniformLocation(progID, "projection");
    modelviewLoc = GLES20.glGetUniformLocation(progID, "modelview");
    normalMatrixLoc = GLES20.glGetUniformLocation(progID, "normalMat");
    modeLoc = GLES20.glGetUniformLocation(progID, "mode");
  
  }
  
  private String loadShaderSrc(int resourceId) {
    StringBuilder sb = new StringBuilder();
    try {
      InputStream is = appContext.getResources().openRawResource(resourceId);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line;
      while ((line = br.readLine()) != null) {
        sb.append(line);
        sb.append('\n');
      }
      br.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return sb.toString();
  }
 
  private void printShaderInfoLog(int obj) {
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    GLES20.glGetShaderiv(obj, GLES20.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    int infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      Log.e(TAG, GLES20.glGetShaderInfoLog(obj));
    }
  }
  
  private void printProgramInfoLog(int obj) {
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    GLES20.glGetProgramiv(obj, GLES20.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    int infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      Log.e(TAG, GLES20.glGetProgramInfoLog(obj));
    }
  }

  private float[] loadVertexData(int resourceId, int perVertexFloats) {

    float[] floatArray = new float[0];
    // read vertex data from file
    int vertSize = 0;
    try {
      InputStream is = appContext.getResources().openRawResource(resourceId);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line = br.readLine();
      if (line != null) {
        vertSize = Integer.parseInt(line);
        floatArray = new float[vertSize];
      }
      int i = 0;
      while ((line = br.readLine()) != null && i < floatArray.length) {
        floatArray[i] = Float.parseFloat(line);
        i++;
      }
      if (i != vertSize || (vertSize % perVertexFloats) != 0) {
        floatArray = new float[0];
      }
      br.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    
    return floatArray;
  }
  
  // the following functions are some matrix and vector helpers
  // they work for this demo but in general it is recommended
  // to use more advanced matrix libraries
  private float vec3Dot(float[] a, float[] b) {
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
  }

  private void vec3Cross(float[] a, float[] b, float[] res) {
    res[0] = a[1] * b[2]  -  b[1] * a[2];
    res[1] = a[2] * b[0]  -  b[2] * a[0];
    res[2] = a[0] * b[1]  -  b[0] * a[1];
  }

  private void vec3Normalize(float[] a) {
    float mag = (float) Math.sqrt(a[0] * a[0]  +  a[1] * a[1]  +  a[2] * a[2]);
    a[0] /= mag; a[1] /= mag; a[2] /= mag;
  } 
  
  private void mat4Identity(float[] a) {
    for (int i = 0; i < 16; ++i) a[i] = 0.0f;
    for (int i = 0; i < 4; ++i) a[i + i * 4] = 1.0f;
  }

  private void mat4Multiply(float[] a, float[] b, float[] res) {
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        res[j*4 + i] = 0.0f;
        for (int k = 0; k < 4; ++k) {
          res[j*4 + i] += a[k*4 + i] * b[j*4 + k];
        }
      }
    }
  }
  
  private void mat4Perspective(float[] a, float fov, float aspect, float zNear, float zFar) {
    float f = 1.0f / (float) (Math.tan (fov/2.0f * (Math.PI / 180.0f)));
    mat4Identity(a);
    a[0] = f / aspect;
    a[1 * 4 + 1] = f;
    a[2 * 4 + 2] = (zFar + zNear)  / (zNear - zFar);
    a[3 * 4 + 2] = (2.0f * zFar * zNear) / (zNear - zFar);
    a[2 * 4 + 3] = -1.0f;
    a[3 * 4 + 3] = 0.0f;
  }
  
  private void mat4LookAt(float[] viewMatrix,
      float eyeX, float eyeY, float eyeZ,
      float centerX, float centerY, float centerZ,
      float upX, float upY, float upZ) {

    float dir[] = new float[3];
    float right[] = new float[3];
    float up[] = new float[3];
    float eye[] = new float[3];

    up[0]=upX; up[1]=upY; up[2]=upZ;
    eye[0]=eyeX; eye[1]=eyeY; eye[2]=eyeZ;

    dir[0]=centerX-eyeX; dir[1]=centerY-eyeY; dir[2]=centerZ-eyeZ;
    vec3Normalize(dir);
    vec3Cross(dir,up,right);
    vec3Normalize(right);
    vec3Cross(right,dir,up);
    vec3Normalize(up);
    // first row
    viewMatrix[0]  = right[0];
    viewMatrix[4]  = right[1];
    viewMatrix[8]  = right[2];
    viewMatrix[12] = -vec3Dot(right, eye);
    // second row
    viewMatrix[1]  = up[0];
    viewMatrix[5]  = up[1];
    viewMatrix[9]  = up[2];
    viewMatrix[13] = -vec3Dot(up, eye);
    // third row
    viewMatrix[2]  = -dir[0];
    viewMatrix[6]  = -dir[1];
    viewMatrix[10] = -dir[2];
    viewMatrix[14] =  vec3Dot(dir, eye);
    // forth row
    viewMatrix[3]  = 0.0f;
    viewMatrix[7]  = 0.0f;
    viewMatrix[11] = 0.0f;
    viewMatrix[15] = 1.0f;
  }
  
  private void mat4Print(float[] a) {
    // opengl uses column major order
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        System.out.print( a[j * 4 + i] + " ");
      }
      System.out.println(" ");
    }
  }

  private void mat4Transpose(float[] a, float[]transposed) {
    int t = 0;
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        transposed[t++] = a[j * 4 + i];
      }
    }
  }

  private boolean mat4Invert(float[] m, float[] inverse) {
    float inv[] = new float[16];
    inv[0] = m[5]*m[10]*m[15]-m[5]*m[11]*m[14]-m[9]*m[6]*m[15]+
             m[9]*m[7]*m[14]+m[13]*m[6]*m[11]-m[13]*m[7]*m[10];
    inv[4] = -m[4]*m[10]*m[15]+m[4]*m[11]*m[14]+m[8]*m[6]*m[15]-
             m[8]*m[7]*m[14]-m[12]*m[6]*m[11]+m[12]*m[7]*m[10];
    inv[8] = m[4]*m[9]*m[15]-m[4]*m[11]*m[13]-m[8]*m[5]*m[15]+
             m[8]*m[7]*m[13]+m[12]*m[5]*m[11]-m[12]*m[7]*m[9];
    inv[12]= -m[4]*m[9]*m[14]+m[4]*m[10]*m[13]+m[8]*m[5]*m[14]-
             m[8]*m[6]*m[13]-m[12]*m[5]*m[10]+m[12]*m[6]*m[9];
    inv[1] = -m[1]*m[10]*m[15]+m[1]*m[11]*m[14]+m[9]*m[2]*m[15]-
             m[9]*m[3]*m[14]-m[13]*m[2]*m[11]+m[13]*m[3]*m[10];
    inv[5] = m[0]*m[10]*m[15]-m[0]*m[11]*m[14]-m[8]*m[2]*m[15]+
             m[8]*m[3]*m[14]+m[12]*m[2]*m[11]-m[12]*m[3]*m[10];
    inv[9] = -m[0]*m[9]*m[15]+m[0]*m[11]*m[13]+m[8]*m[1]*m[15]-
             m[8]*m[3]*m[13]-m[12]*m[1]*m[11]+m[12]*m[3]*m[9];
    inv[13]= m[0]*m[9]*m[14]-m[0]*m[10]*m[13]-m[8]*m[1]*m[14]+
             m[8]*m[2]*m[13]+m[12]*m[1]*m[10]-m[12]*m[2]*m[9];
    inv[2] = m[1]*m[6]*m[15]-m[1]*m[7]*m[14]-m[5]*m[2]*m[15]+
             m[5]*m[3]*m[14]+m[13]*m[2]*m[7]-m[13]*m[3]*m[6];
    inv[6] = -m[0]*m[6]*m[15]+m[0]*m[7]*m[14]+m[4]*m[2]*m[15]-
             m[4]*m[3]*m[14]-m[12]*m[2]*m[7]+m[12]*m[3]*m[6];
    inv[10]= m[0]*m[5]*m[15]-m[0]*m[7]*m[13]-m[4]*m[1]*m[15]+
             m[4]*m[3]*m[13]+m[12]*m[1]*m[7]-m[12]*m[3]*m[5];
    inv[14]= -m[0]*m[5]*m[14]+m[0]*m[6]*m[13]+m[4]*m[1]*m[14]-
             m[4]*m[2]*m[13]-m[12]*m[1]*m[6]+m[12]*m[2]*m[5];
    inv[3] = -m[1]*m[6]*m[11]+m[1]*m[7]*m[10]+m[5]*m[2]*m[11]-
             m[5]*m[3]*m[10]-m[9]*m[2]*m[7]+m[9]*m[3]*m[6];
    inv[7] = m[0]*m[6]*m[11]-m[0]*m[7]*m[10]-m[4]*m[2]*m[11]+
             m[4]*m[3]*m[10]+m[8]*m[2]*m[7]-m[8]*m[3]*m[6];
    inv[11]= -m[0]*m[5]*m[11]+m[0]*m[7]*m[9]+m[4]*m[1]*m[11]-
             m[4]*m[3]*m[9]-m[8]*m[1]*m[7]+m[8]*m[3]*m[5];
    inv[15]= m[0]*m[5]*m[10]-m[0]*m[6]*m[9]-m[4]*m[1]*m[10]+
             m[4]*m[2]*m[9]+m[8]*m[1]*m[6]-m[8]*m[2]*m[5];

    float det = m[0]*inv[0]+m[1]*inv[4]+m[2]*inv[8]+m[3]*inv[12];
    if (det == 0) return false;
    det = 1.0f / det;
    for (int i = 0; i < 16; i++) inverse[i] = inv[i] * det;
    return true;
  }

}

class GLSurfaceViewRenderer implements GLSurfaceView.Renderer {

  private Renderer renderer;
  
  final int targetFPS = 60;
  final int targetMillis = 1000 / targetFPS;
  private long startTime = 0;
  
  public GLSurfaceViewRenderer(Context appContext) {
    renderer = new Renderer(appContext);
  }
  
  @Override
  public void onDrawFrame(GL10 gl10) {
    float offset = 1.0f;
    renderer.t += offset;
    renderer.display(gl10);
    
    // this code attempts to achieve a target frames-per-second rate
    long timeTillNextRender = startTime + targetMillis;
    long sleepTime = timeTillNextRender - SystemClock.uptimeMillis();
    if (sleepTime > 0) {
      try {
        Thread.sleep(sleepTime);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    } else {
      // rendering is too slow
    }
    startTime = SystemClock.uptimeMillis();
  }

  @Override
  public void onSurfaceChanged(GL10 gl10, int width, int height) {
    renderer.resize(gl10, width, height);
  }

  @Override
  public void onSurfaceCreated(GL10 gl10, EGLConfig config) {
    renderer.init(gl10, config);
    startTime = SystemClock.uptimeMillis();
  }
  
  public Renderer getRenderer() {
    return renderer;
  }
}

public class ShaderLightMat extends Activity {

  private GLSurfaceView myGLSurfaceView;
  private GLSurfaceViewRenderer myGLSurfaceViewRenderer;
  private GestureDetector gestDetector;
  
  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    myGLSurfaceView = new GLSurfaceView(this);
    
    // Create an OpenGL ES 2.0 context.
    myGLSurfaceView.setEGLContextClientVersion(2);
    
    myGLSurfaceViewRenderer = new GLSurfaceViewRenderer(this);
    myGLSurfaceView.setRenderer(myGLSurfaceViewRenderer);
    // myGLSurfaceView.setRenderMode(GLSurfaceView.RENDERMODE_WHEN_DIRTY);
    setContentView(myGLSurfaceView);
    setTitle("Fling to change model/shader");
    MyGestureListener gl = new MyGestureListener();
    gestDetector = new GestureDetector(this, gl);
  }
  
  @Override
  protected void onPause() {
    // Large objects in graphics memory should be de-allocated here
    // This is skipped here for simplicity of the example.
    super.onPause();
    myGLSurfaceView.onPause();
  }

  @Override
  protected void onResume() {
    super.onResume();
    // Large objects in graphics memory should be re-created here
    // if they were de-allocated on onPause().
    // This is skipped here for simplicity of the example.
    myGLSurfaceView.onResume();
  }
  
  public boolean onTouchEvent(MotionEvent e) {
    this.gestDetector.onTouchEvent(e);
    return super.onTouchEvent(e);
  }
  
  class MyGestureListener extends GestureDetector.SimpleOnGestureListener {

    @Override
    public boolean onDown(MotionEvent e) {      

      return true;
    }

    @Override
    public void onLongPress(MotionEvent e) {
      int mode = myGLSurfaceViewRenderer.getRenderer().modeVal;
      mode ++;
      if(mode > 5) {
        mode = 1;
      }
      String modeStr = "Mode = " + mode;
      setTitle(modeStr);
      myGLSurfaceViewRenderer.getRenderer().modeVal = mode;
    }

    @Override
    public boolean onFling(MotionEvent e1, MotionEvent e2, float velocityX, float velocityY) {
      Renderer r = myGLSurfaceViewRenderer.getRenderer();
      
      if(Math.abs(velocityX) > Math.abs(velocityY)) {
        // fling left/right changes model
        r.loadNext = true;
        if(velocityX > 0) {
          r.vboFile++;
        }else{
          r.vboFile--;
        }
        if(r.vboFile > 4) r.vboFile = 0;
        if(r.vboFile < 0) r.vboFile = 4;
        
        if(r.vboFile == 0) r.vboResourceId = R.raw.knot;
        if(r.vboFile == 1) r.vboResourceId = R.raw.teapot;
        if(r.vboFile == 2) r.vboResourceId = R.raw.sphere;
        if(r.vboFile == 3) r.vboResourceId = R.raw.hose;
        if(r.vboFile == 4) r.vboResourceId = R.raw.cube;
       
      }else{
        // fling top/down changes shader
        String  modeStr = "Shader Light/Material";
        r.loadNext = true;
        if(velocityY > 0) {
          r.shaderFile++;
        }else{
          r.shaderFile--;
        }
        if(r.shaderFile > 3) r.shaderFile = 0; 
        if(r.shaderFile < 0) r.shaderFile = 3; 
        if(r.shaderFile == 0) {
          r.shaderVertId = R.raw.phongbrdfpervertex_vert;
          r.shaderFragId = R.raw.phongbrdfpervertex_frag;
          modeStr = "PhongBrdfPerVertex";
        }
        if(r.shaderFile == 1) {
          r.shaderVertId = R.raw.phongbrdfperfrag_vert;
          r.shaderFragId = R.raw.phongbrdfperfrag_frag;
          modeStr = "PhongBrdfPerFrag";
        }
        if(r.shaderFile == 2) {
          r.shaderVertId = R.raw.phongshading_vert;
          r.shaderFragId = R.raw.phongshading_frag;
          modeStr = "PhongShading";
        }
        if(r.shaderFile == 3) {
          r.shaderVertId = R.raw.blinnphongshading_vert;
          r.shaderFragId = R.raw.blinnphongshading_frag;
          modeStr = "BlinnPhongShading";
        }
        setTitle(modeStr);
      }
      return true;
    }
  }
}
